from .misc import multiton


class ClumpSet(set):
    """
    Holds a set of clump-checkers (and facilitates checking therewith)
    """
    def successes(self, parsed):
        """
        return: all entity names that are or are not in `parsed` as expected
        """
        return {name for c in self if c.verify(parsed) for name in c.member_names}
    
    def failures(self, parsed):
        """
        return: generator of (expected, the_only_successes) from all clumps that
        `parsed` does not satisfy
        """
        return ((c.member_names, c.to_eliminate(parsed)) for c in self if not c.verify(parsed))


class _Clump:
    def __init__(self, key, host):
        """
        key: Unique string with which to group other ANDs
        host: Handler creating this clump; only used for multiton
          instance-checking
        """
        self.key = key
        self.host = host
        # self.members holds everything this clump checks against
        self.members = set()
    
    @property
    def member_names(self):
        return frozenset(i.name for i in self.members)
    
    def add(self, item):
        self.members.add(item)


@multiton()
class And(_Clump):
    def verify(self, parsed):
        # Should contain either no members or all members (latter indicating none were given)
        diff = self.member_names.difference(parsed)
        return not diff or parsed == diff
    
    def to_eliminate(self, parsed):
        """
        return: entities in `parsed` that were expected to be present.
        (that is, all these methods return successes)
        """
        return frozenset(self.member_names.intersection(parsed))


@multiton()
class Or(_Clump):
    def verify(self, parsed):
        # Should contain at least 1 member
        return bool(self.member_names.intersection(parsed))
    
    def to_eliminate(self, parsed):
        """
        return: entities in `parsed` that were expected to be present.
        """
        return frozenset(self.member_names.intersection(parsed))


@multiton()
class Xor(_Clump):
    def verify(self, parsed):
        # Should contain exactly 1 member
        return len(self.member_names.intersection(parsed)) == 1
    
    def to_eliminate(self, parsed):
        """
        return: entities not in `parsed` that were not expected to be present.
        """
        return frozenset(self.member_names.difference(parsed))



from .misc import JoffreyNamespace


class JoffreyException(Exception):
    """
    Base class for joffrey-related exceptions; includes a
    "details" attribute for whatever error-related info.
    """
    def __init__(self, msg, **kwargs):
        self.details = JoffreyNamespace(**kwargs)
        super().__init__(msg)


class ANDError(JoffreyException):
    pass


class ORError(JoffreyException):
    pass


class XORError(JoffreyException):
    pass


class RequirementError(JoffreyException):
    pass



"""
argparse sucks
this sucks too but less
"""
import os
import sys
import shlex
from functools import partial
from itertools import chain, zip_longest, starmap

from . import errors
from .clumps import And, Or, Xor, ClumpSet
from .entities import Entity, Arg, Flag
from .misc import JoffreyNamespace, _Null


_FILE = os.path.basename(sys.argv[0])


class HelperMixin:
    """
    Provides help-screen functionality to Handlers.
    A bit convoluted; needs reworking.
    """
    
    @property
    def all_commands(self):
        return {*self.commands, *(name for g in self._groups for name in g.commands)}
    
    @property
    def all_flags(self):
        return self.flags.values()
    
    @property
    def all_args(self):
        return map(self.getarg, self.args)
    
    @property
    def usage_info(self):
        return '{}{} {} {}'.format(
          _FILE,
          str(self),
          ' '.join(map(str, self.all_flags)),
          ' '.join(map(str, self.all_args)),
          )
    
    @property
    def help_info(self):
        return '{}\n{}'.format(
          self._label_format('args'),
          self._label_format('flags')
          )
    
    def _label_format(self, label):
        return '{}\n{}'.format(
          label.upper(),
          '\n'.join({
            '\t{: <15} {}'.format(i.name, i.brief)
            for i in getattr(self, 'all_' + label)
            })
          )
    
    def format_help(self, usage=True, commands=True, help=True):
        built = ['', self.desc, ''] if self.desc else ['']
        if usage:
            built.append('usage: {}'.format(self.usage_info))
        if commands and self.commands:
            built.append('commands: {}'.format(', '.join(map(str, self.all_commands))))
        if help:
            if usage or commands:
                built.append('')
            built.append(self.help_info)
        return '\n'.join(built)
    
    def print_help(self, usage=True, commands=True, help=True):
        print(self.format_help(usage, commands, help), end='\n\n')
    
    def error(self, exc=None, help=True):
        self.print_help(help=help)
        if exc is None:
            raise SystemExit
        raise SystemExit(exc if str(exc) else type(exc))
    
    def cli_help(self, name=None):
        """
        name: name of entity to provide help for
        If no name given, provides general help
        """
        if name is None:
            # (this prints the help screen)
            self.error()
        
        entity = self.get(name)
        if entity is None:
            print('No helpable entity named', repr(name))
            raise SystemExit
        
        short = getattr(entity, 'short', '')
        try:
            aliases = ', '.join(map(repr, (k for k, v in self._aliases.items() if v == entity.name and k != short)))
        except AttributeError:
            aliases = ''
        
        name = str(entity).lstrip()
        if aliases:
            print('', name, 'aliases: {}'.format(aliases), entity.help, sep='\n')
        else:
            print('', name, entity.help, sep='\n')
        raise SystemExit


class _Handler:
    """
    Base class for anything that 'handles' flags/args.

    _aliases: Dict of {alias: entity's real name}
    _defaults: Dict of {entity name: default value if not called}
      which is used to fill in unprovided values after parsing
    _last_arg_consumes: Whether the last provided positional argument
      has a repeat count of "...", aka infinite
    _and: AND clumps to take into account when parsing
    _or: OR clumps
    _xor: XOR clumps
    _required: Set of all entities created with required=True
      (for which to raise an error if not provided)
    
    arg_map: Dict of {arg name: joffrey.entity.Arg object}
    commands: Dict of {command name: joffrey.core.Command object}
    flags: Dict of {flag name: joffrey.entity.Flag object}
    args: List of positional arguments provided in the current run
    """

    def __init__(self):
        self.arg_map = {}
        self.commands = {}
        self.flags = {}
        self._aliases = {}
        self._defaults = {}
        
        self.args = []
        self._last_arg_consumes = False
        
        self._and = ClumpSet()
        self._or = ClumpSet()
        self._xor = ClumpSet()
        
        self._required = set()
    
    def __repr__(self):
        quote = "'" if hasattr(self, 'name') else ''
        return '<{}{s}{q}{}{q}>'.format(
          self.__class__.__name__,
          getattr(self, 'name', ''),
          s=' ' if hasattr(self, 'name') else '',
          q=quote,
          )
    
    @property
    def defaults(self):
        # All subcommands would have a default value of their own defaults, hence the dict comp
        return JoffreyNamespace(**self._defaults, **{cmd.name: cmd.defaults for cmd in self.commands.values()})
    
    @property
    def parent_and(self):
        # These are defined differently in subclasses
        return self._and
    
    @property
    def parent_or(self):
        return self._or
    
    @property
    def parent_xor(self):
        return self._xor
    
    @property
    def entity_names(self):
        return {*self.arg_map, *self.commands, *self.flags}
    
    def dealias(self, name):
        return self._aliases.get(name, name)
    
    def remove(self, obj):
        name = obj.identifier if isinstance(obj, Entity.cls) else self.dealias(obj)
        if name in self.arg_map:
            if self._last_arg_consumes:
                # If the entity being removed *is* the one that consumes, then
                # _last_arg_consumes can no longer be True
                self._last_arg_consumes = self.arg_map[name].repcount is not Ellipsis
            self.args = [n for n in self.args if n != name]
            del self.arg_map[name]
        elif name in self.commands:
            del self.commands[name]
        elif name in self.flags:
            del self.flags[name]
        else:
            raise KeyError('No such entity: {}'.format(obj))
        self._aliases = {k: v for k, v in self._aliases.items() if v != name}
    
    def get(self, name):
        """
        Generic getter; use getarg/getflag/getcmd for specific type of entity
        (analogous to hasany())
        """
        for func in (self.getarg, self.getflag, self.getcmd):
            try:
                return func(name)
            except KeyError:
                pass
        return None
    
    def getarg(self, name):
        try:
            return self.arg_map[name]
        except KeyError:
            return self.arg_map[self._aliases[name]]
    
    def getflag(self, name):
        try:
            return self.flags[name]
        except KeyError:
            return self.flags[self._aliases[name]]
    
    def getcmd(self, name):
        try:
            return self.commands[name]
        except KeyError:
            return self.commands[self._aliases[name]]
    
    def hasflag(self, name):
        return self.dealias(name) in self.flags
    
    def hascmd(self, name):
        return self.dealias(name) in self.commands
    
    def hasany(self, name):
        # Analogous to get()
        return self.hasflag(name) or self.hascmd(name) or self.dealias(name) in self.arg_map
    
    def _clump(self, obj, AND, OR, XOR):
        """
        AND/OR/XOR: a clump's unique identifier
        obj: Object to clump with any of the above
        
        Clump object will be grabbed according to
        identifier, and obj will be added into it
        """
        if AND is not _Null:
            self._and.add(And(AND, self))
            And(AND, self).add(obj)
        if OR is not _Null:
            self._or.add(Or(OR, self))
            Or(OR, self).add(obj)
        if XOR is not _Null:
            self._xor.add(Xor(XOR, self))
            Xor(XOR, self).add(obj)
    
    def enforce_clumps(self, parsed, groups=None):
        """
        parsed: Set of entities' names that were extracted from user input
        groups: Clump-groups to take into account when checking
        
        Enforce AND/OR/XOR rules. Mostly the 'heart' of joffrey.
        """
        
        # Entities to eliminate for each clump
        # (elimination means it was receieved as expected)
        elim = {
          'AND': self.parent_and.successes(parsed),
          'OR': self.parent_or.successes(parsed),
          'XOR': self.parent_xor.successes(parsed)
          }
        
        def extract_names(collection):
            """
            Only used if no groups were given to enforce_clumps()
            (which means there's no nesting to take care of, which
            in turn means a simple repr over everything will do)
            """
            return map(repr, collection)
        
        if groups is not None:
            def extract_names(collection):
                """[Visually group together] group's entity names"""
                return (
                  '[{}]'.format(', '.join(map(repr, groups[n].entity_names)))
                  if n in groups else repr(n) for n in collection
                  )
            _g_clumps = {lbl: {groups.get(s) for s in successes.intersection(groups)} for lbl, successes in elim.items()}
            elim = {lbl: elim[lbl] | {name for g in grps for name in g.entity_names} for lbl, grps in _g_clumps.items()}
        
        err_details = {
          'parsed': parsed,
          'groups': groups,
          'handler': repr(self),
          'AND_SUC': elim['AND'],  # SUC = SUCCESSES
          'OR_SUC': elim['OR'],
          'XOR_SUC': elim['XOR'],
          }
        
        for all_failed, received in self._and.failures(parsed):
            # AND failure == member of an AND clump that was not given
            # an AND failure is okay if it's in a satisfied OR clump (i.e. there's at least one other OR in its clump that was given)
            # or if it's in a satisfied XOR clump (i.e. exactly one other XOR in its clump was given)
            not_exempt = (all_failed - received) - elim['OR'] - elim['XOR']
            if not_exempt:
                raise errors.ANDError(
                  'Expected all of the following flags/arguments/commands: {}\n(Got {})'.format(
                      ', '.join(extract_names(all_failed)),
                      ', '.join(extract_names(received)) or 'none'
                    ),
                  **err_details,
                  failed=all_failed, eliminating=received, not_exempt=not_exempt
                  )
        
        for all_failed, received in self._or.failures(parsed):
            # OR failure == member of an OR clump where none were given
            # an OR failure is okay if it's in a satisfied XOR clump (i.e. exactly one other XOR in its clump was given)
            not_exempt = (all_failed - received) - elim['XOR']
            if not_exempt:
                raise errors.ORError(
                  'Expected at least one of the following flags/arguments/commands: {}\n(Got none)'.format(
                      ', '.join(extract_names(all_failed))
                    ),
                  **err_details,
                  failed=all_failed, eliminating=received, not_exempt=not_exempt
                  )
        
        for all_failed, not_received in self._xor.failures(parsed):
            # XOR failure == member of an XOR clump that was given alongside at least one other
            # an XOR failure is okay if it satisfies an AND clump (i.e. all other ANDs in its clump were given)
            ######################################################################################
            #  XXX: SHOULD it be okay if it satisfies an OR clump? What about if it's required?  #
            #  That is: should this actually say `- elim['AND'] - elim['OR'] - self._required`?  #
            ######################################################################################
            not_exempt = (all_failed - not_received) - elim['AND']
            if len(not_exempt) > 1:
                raise errors.XORError(
                  'Expected no more than one of the following flags/arguments/commands: {}\n(Got {})'.format(
                      ', '.join(extract_names(all_failed)),
                      ', '.join(extract_names(all_failed-not_received))
                    ),
                  **err_details,
                  failed=all_failed, eliminating=not_received, not_exempt=not_exempt
                  )
        return True
    
    def clump(self, *, AND=_Null, OR=_Null, XOR=_Null):
        """
        AND/OR/XOR: Identifiers for each clump
        return: Wrapper function

        Decorator that proxies _clump(), adding an entity
        to the requested clumps
        """
        def inner(cb):
            entity = Entity(cb.func if isinstance(cb, Entity.cls) else cb)
            self._clump(entity, AND, OR, XOR)
            return entity
        return inner
    
    def arg(self, n=1, *, required=False, default=_Null, namespace=None, help=None):
        """
        n: number of times this arg should be received consecutively; ... for infinite
        required: Whether this arg is required to be provided
        default: Default value if this arg is not provided
        namespace: Dict with which to initialize Arg entity's namespace (for storing state between repeated calls)
        help: Help text for this arg (preferably None: defaults to function's __doc__)

        Decorator that registers its decorated function as a positional argument
        """
        def inner(cb):
            repeat_count = n
            entity = Arg(cb, n, namespace=namespace, help=help)
            self.arg_map[entity.name] = entity
            if required:
                self._required.add(entity.name)
            if repeat_count is Ellipsis:
                self._last_arg_consumes = True
                repeat_count = 1
            if default is not _Null:
                self._defaults[entity.name] = default
            self.args.extend([entity.name] * repeat_count)
            return entity
        return inner
    
    def flag(self, dest=None, short=_Null, *, aliases=(), required=False, default=_Null, namespace=None, help=None, _='-'):
        """
        dest: What this flag's name should be in the final resultant JoffreyNamespace
        short: Shorthand alias for this flag; _Null => first available letter, None = no short alias
        aliases: Other aliases for this flag
        required: Whether this flag is required to be provided
        default: Default value if this flag is not provided
        namespace: Dict with which to initialize Flag entity's namespace (for storing state between repeated calls)
        help: Help text for this flag (preferably None: defaults to function's __doc__)

        Decorator that registers its decorated function as a flag/option
        """
        def inner(cb):
            entity = Flag(cb, namespace=namespace, name=dest, help=help, _=_)
            # filter out '<lambda>'
            if cb.__name__.isidentifier():
                self._aliases[cb.__name__] = entity.name
            if short is not None:  # _Null == default; None == none
                try:
                    entity.short = short or next(s for s in entity.name if s.isalnum() and s not in self._aliases)
                except StopIteration:
                    pass
                else:
                    self._aliases[entity.short] = entity.name
            if default is not _Null:
                self._defaults[entity.identifier] = default
            if required:
                self._required.add(entity.name)
            self._aliases.update(dict.fromkeys(aliases, entity.name))
            self.flags[entity.name] = entity
            return entity
        return inner
    
    def command(self, name, desc='', *args, AND=_Null, OR=_Null, XOR=_Null, from_cli=None, aliases=(), _='-', **kwargs):
        """
        name: This command's name
        desc: This command's helptext (a short description of it)
        AND/OR/XOR: Clump values for this command as a whole
        from_cli: If not None, joffrey.core.CLI object to create this command from
        aliases: Aliases for this command
        _: What to replace underscores in this command's name with ((XXX: is that even applicable? commands aren't attrs))
        *args, **kwargs: See Command.__init__()

        Creates and returns a Command object that acts as a sub-command/sub-parser
        """
        if from_cli is None:
            subcmd = Command(*args, **kwargs, name=name, desc=desc, parent=self)
        else:
            subcmd = Command.from_cli(from_cli, self, name)
        visual_name = name.replace('_', _)
        self._aliases.update(dict.fromkeys(aliases, visual_name))
        self._aliases[name] = visual_name
        self.commands[visual_name] = subcmd
        self._clump(subcmd, AND, OR, XOR)
        return subcmd


class ParserBase(_Handler, HelperMixin):
    """
    Base class for Handlers that can parse arguments.

    _groups: Clump-groups present in this parser
    _prepared_parse: A functools.partial that is set once prepare() is called
    _result: The result of _prepared_parse(), once called
    
    desc: Helptext (a short description of this parser)
    flag_prefix: What to prefix shorthand flags with
    long_prefix: flag_prefix*2, used to prefix long-form flags
    systemexit: Whether to raise SystemExit on error or just fail w/ the original exception
    """
    
    def __init__(self, desc='', flag_prefix='-', *, systemexit=True, no_help=False):
        """
        desc: Helptext (a short description of this parser)
        flag_prefix: What to prefix shorthand flags with (long prefix is this*2)
        systemexit: Whether to raise SystemExit on error or just fail w/ the original exception
        no_help: Whether NOT to create a default help command
        """
        super().__init__()
        self.desc = desc
        self.flag_prefix = flag_prefix
        self.long_prefix = 2 * flag_prefix
        self.systemexit = systemexit
        self._groups = set()
        self._prepared_parse = None
        self._result = None
        if not flag_prefix:
            raise ValueError('Flag prefix cannot be empty')
        if not no_help:
            self.flag('help',
              help="Prints help and exits\nIf given a valid NAME, displays that entity's help"
            )(lambda name=None: self.cli_help(name))
    
    def __setattr__(self, name, val):
        """
        Special-cased for joffrey.core.Group objects: re-initializes the
        group as a SubHandler and adds it to self._groups
        """
        if not isinstance(val, Group):
            return object.__setattr__(self, name, val)
        
        if name in vars(self):
            raise ValueError('Group name already in use for this cli: ' + name)
        if val._required:
            self._required.add(name)
        
        self._groups.add(val)
        self._clump(val, val._and, val._or, val._xor)
        
        SubHandler.__init__(val, self, name)
        object.__setattr__(self, name, val)
    
    @property
    def defaults(self):
        return super().defaults
    
    @property
    def result(self):
        """
        Returns the result of self._prepared_parse()
        once that is set to not-None.
        In the grand scheme of things, allows projects to be based
        entirely around the CLI (by getting default-or-passed values
        from cli.result).
        """
        if self._prepared_parse is None:
            return self.defaults
        if self._result is None:
            self._result = self._prepared_parse()
        return self._result
    
    def dealias(self, name):
        try:
            return next(g.dealias(name) for g in self._groups if g.hasany(name))
        except StopIteration:
            return super().dealias(name)
    
    def remove(self, obj):
        name = obj.name if isinstance(obj, Entity.cls) else obj
        for g in self._groups:
            try:
                return g.remove(name)
            except KeyError:
                pass
        return super().remove(name)
    
    def get(self, name):
        try:
            return next(g.get(name) for g in self._groups if g.hasany(name))
        except StopIteration:
            return super().get(name)
    
    def getarg(self, name):
        try:
            return next(g.getarg(name) for g in self._groups if name in g.arg_map)
        except StopIteration:
            return super().getarg(name)
    
    def getflag(self, name):
        try:
            return next(g.getflag(name) for g in self._groups if g.hasflag(name))
        except StopIteration:
            return super().getflag(name)
    
    def getcmd(self, name):
        try:
            return next(g.getcmd(name) for g in self._groups if g.hascmd(name))
        except StopIteration:
            return super().getcmd(name)
    
    def hasflag(self, name):
        return super().hasflag(name) or any(g.hasflag(name) for g in self._groups)
    
    def hascmd(self, name):
        return super().hascmd(name) or any(g.hascmd(name) for g in self._groups)
    
    def hasany(self, name):
        # XXX: What reason does this have for not checking `g.hasany() for g in self._groups`?
        # (is it intentional? I don't remember... my bad for not commenting as soon as I wrote it.)
        return super().hasany(name) or self._aliases.get(name, _Null) in self.arg_map
    
    def enforce_clumps(self, parsed):
        # p is a set of `parsed` PLUS names of all groups that have entities in `parsed`
        # because the group names are what this parser's and/or/xor clumps will be looking for
        p = set(parsed).union(next((g.name for g in self._groups if g.hasany(i)), None) for i in parsed) - {None}
        return (
          super().enforce_clumps(p, {g.name: g for g in self._groups})
          and
          all(g.enforce_clumps(parsed) for g in self._groups if g.name in p)
          )
    
    def _put_nsp(self, namespaces, entity):
        """
        namespaces: dict of namespaces used in current parse session
        entity: entity to pass a namespace to
        return: Callable that provides entity with the appropriate namespace

        Since namespaces are local to each parse-session, they can
        be stored in a single dict (viz. namespaces) of {entity: nsp}.
        """
        return entity if entity.namespace is None else partial(
          entity,
          namespaces.setdefault(
            entity.name,
            JoffreyNamespace(**entity.namespace)
            )
          )
    
    def _check_skip(self, value):
        """
        Check if this value ends a run of flag arguments
        (i.e. if it's another flag or the end-of-flags sentinel)
        """
        return value.startswith(self.flag_prefix) or value == '--'
    
    def _extract_flargs(self, inp, strict=False, propagate_unknowns=False):
        """
        inp: Input to parse
        strict: Whether to disallow unknown flags / excessive positional args
        propagate_unknowns: Whether unknown flags should be regarded as
          an error or just bubbled up to parent handler
        return: flags+args found, subcommand (if any), and unknown flags to propagate
        
        Extract flags/args from user input and return both.
        """
        # args/flags found so far
        flags = []
        args = []
        # number of elements to skip (used when flag has multiple args)
        skip = 0
        # subcommand, if any
        command = None
        # start off assuming there can be flags in the input
        # (changes if we hit a '--')
        allow_flags = True
        
        # errors
        too_many_args = False
        unknown_flags = []
        
        for idx, value in enumerate(inp, 1):
            if skip > 0:
                skip -= 1
                continue
            
            # "No flags beyond this point" sentinel
            if value == '--':
                allow_flags = False
                continue
            
            if (not allow_flags) or (not value.startswith(self.flag_prefix) or value in (self.flag_prefix, self.long_prefix)):
                # Then value is a command or positional argument
                if self.hascmd(value):
                    command = (value, idx)
                    # Commands consume everything to their right, so no point parsing further
                    break
                args.append(value)
                # self._last_arg_consumes == infinite args allowed
                if not self._last_arg_consumes and len(args) > len(self.args):
                    too_many_args = True
            elif allow_flags:
                if '=' in value:
                    # Then it's passing a single arg to the flag
                    name, arg = value.lstrip(self.flag_prefix).split('=', 1)
                    if self.hasflag(name):
                        flags.append((self.dealias(name), [arg] if arg else []))
                    elif propagate_unknowns:
                        unknown_flags.append(('', value.split('=')[0], [arg] if arg else []))
                    else:
                        unknown_flags.append(('', value.split('=')[0]))
                    continue
                
                if value.startswith(self.long_prefix):
                    name = value.lstrip(self.flag_prefix)
                    if self.hasflag(name):  # long-form flag name
                        skip = self.getflag(name).argcount
                        next_pos = next((i for i, v in enumerate(inp[idx:]) if self._check_skip(v)), len(inp))
                        if next_pos < skip:
                            skip = next_pos
                        flags.append((self.dealias(name), inp[idx:skip+idx]))
                    elif propagate_unknowns:
                        # Below is commented out because there's no way of knowing how many args the flag accepts
                        # if it's not this parser's own
                        #skip = next((i for i, v in enumerate(inp[idx:]) if self._check_skip(v)), len(inp))
                        #unknown_flags.append(('', value, inp[idx:skip+idx]))
                        unknown_flags.append(('', value, []))
                    else:
                        unknown_flags.append(('', value))
                    continue
                
                for name in value[1:]:  # collection of shorthand flag names (like '-xcvf')
                    if self.hasflag(name):
                        skip = self.getflag(name).argcount
                        next_pos = next((i for i, v in enumerate(inp[idx:]) if self._check_skip(v)), len(inp))
                        if next_pos < skip:
                            skip = next_pos
                        flags.append((self.dealias(name), inp[idx:skip+idx]))
                    elif propagate_unknowns:
                        unknown_flags.append((value[0], name, []))
                    else:
                        unknown_flags.append((value[0], name))
        
        if strict:
            if too_many_args:
                if self.commands and not self.args:
                    raise TypeError(
                      'Expected a command: {}\n'
                      'Try `--help <command name>` for specific detail'
                      .format(', '.join(map(repr, self.commands)))
                      )
                raise TypeError('Too many positional arguments (expected {}, got {})'.format(
                  len(self.args), len(args)
                  ))
            if unknown_flags and not propagate_unknowns:
                raise TypeError('Unknown flag(s): ' + ' '.join(starmap("`{}{}'".format, unknown_flags)))
        return flags, args, command, unknown_flags if propagate_unknowns else []
    
    def do_parse(self, inp=None, strict=False, systemexit=True, propagate_unknowns=False):
        """
        inp: List of command-line args to parse
        strict: Whether to disallow excessive args and/or unknown non-propagable flags
        systemexit: Whether to raise SystemExit on error or just fail with the original exception
        propagate_unknowns: Whether to bubble up unknown flags to parent handler
        return: Parsed-out JoffreyNamespace from inp, unknown flags found
        
        Backend to parse() -- does the actual parsing and returns result + unknown flags to propagate
        """
        parsed = {}
        flags, positionals, command, unknown_flags = self._extract_flargs(inp, strict, propagate_unknowns)
        # Namespaces to be passed to entities in current run
        # (nsps are parse-session-local)
        namespaces = {}
        prep = partial(self._put_nsp, namespaces)
        
        for flag, args in flags:
            if self.hasflag(flag):
                entity = self.getflag(flag)
                parsed[entity.identifier] = prep(entity)(*args)
        
        if self._last_arg_consumes and len(positionals) > len(self.args):
            # Fill zipped_args with the Arg that's meant to consume trailing values
            zipped_args = zip_longest(map(self.getarg, self.args), positionals, fillvalue=self.getarg(self.args[-1]))
        else:
            zipped_args = zip(map(self.getarg, self.args), positionals)
        
        for entity, value in zipped_args:
            parsed[entity.identifier] = prep(entity)(value)
        
        if command is not None:
            value, idx = command
            command, cmd_obj = self._aliases.get(value, value), self.getcmd(value)
            try:
                parsed[command], cmd_unknown_flags = cmd_obj.do_parse(inp[idx:], strict, systemexit, propagate_unknowns)
            except Exception as e:
                if systemexit is None and cmd_obj.systemexit or systemexit:
                    cmd_obj.error(e, help=False)
                raise
            else:
                if propagate_unknowns:
                    # The _ is the flag's name, which would only have been used for error output
                    for _, flag, args in cmd_unknown_flags:
                        name = flag.lstrip(self.flag_prefix)
                        if self.hasflag(name):
                            entity = self.getflag(name)
                            parsed[entity.identifier] = prep(entity)(*args)
                        else:
                            # Propagate yet further
                            unknown_flags.append((None, name, args))
        self.enforce_clumps(parsed)
        # Place defaults first then override them with provided values
        final = {**self._defaults, **{name: value for g in self._groups for name, value in g._defaults.items()}, **parsed}
        
        nsp = JoffreyNamespace(**final)
        # One final check after enforce_clumps: all required entities must have been provided
        if self._required.difference(nsp):
            raise errors.RequirementError('Expected the following required arguments: {}\nGot {}'.format(
              ', '.join(map(repr, self._required)),
              ', '.join(map(repr, self._required.intersection(nsp))) or 'none'
              ))
        return nsp, unknown_flags
    
    def parse(self, inp=None, *, systemexit=None, strict=False, propagate_unknowns=False):
        """
        inp: Input to parse, either as a string (which is then shlex.split) or a list of string arguments
        systemexit: Whether to raise SystemExit on error or just fail with the original exception
        strict: Whether to disallow excessive args and/or unknown non-propagable flags
        propagate_unknowns: Whether to bubble up unknown flags to parent handler
        return: Resultant JoffreyNamespace from parse

        Parses user input into an JoffreyNamespace. If systemexit, prints usage info on error.
        """
        if inp is None:
            inp = sys.argv[1:]
        if isinstance(inp, str):
            inp = shlex.split(inp)
        
        try:
            nsp, _ = self.do_parse(list(inp), strict, systemexit, propagate_unknowns)  # list(inp) is effectively inp.copy()
        except Exception as e:
            if systemexit is None and self.systemexit or systemexit:
                self.error(e, help=False)
            raise
        except SystemExit:
            if systemexit is None and self.systemexit or systemexit:
                raise
        else:
            return nsp
    
    def prepare(self, *args, **kwargs):
        """
        *args, **kwargs: see parse().
        return: self
        
        Makes self.result return an JoffreyNamespace of actual
        parsed values (not just defaults).
        Returns self to allow a `prepare()` call to be chained
        into `set_defaults()` and/or `result`.
        """
        self._prepared_parse = partial(self.parse, *args, **kwargs)
        return self
    
    def set_defaults(self, **kwargs):
        """
        **kwargs: names and values to insert into self._defaults

        Ensures that each name in kwargs is already present.
        Returns self to allow a `set_defaults()` call to be chained
        into `result` or `prepare()`.
        """
        for name in kwargs:
            if not self.hasany(name):
                raise KeyError('Unknown name {} passed to set_defaults()'.format(name))
        self._defaults.update(kwargs)
        return self


class SubHandler(_Handler):
    """
    Base class for handlers that cannot parse, but are 'attached' to handlers that can.
    """
    def __init__(self, parent, name):
        if not name:
            raise ValueError('Sub-handler name cannot be empty')
        self.name = name
        self.parent = parent
        super().__init__()
    
    @property
    def parent_and(self):
        return ClumpSet(self._aliases.get(i, i) for i in chain(self._and, self.parent.parent_and))
    
    @property
    def parent_or(self):
        return ClumpSet(self._aliases.get(i, i) for i in chain(self._or, self.parent.parent_or))
    
    @property
    def parent_xor(self):
        return ClumpSet(self._aliases.get(i, i) for i in chain(self._xor, self.parent.parent_xor))


class Group(SubHandler):
    """
    Clump-group that hosts clumps within itself but can be itself passed to clumps whole.
    Strange implementation to streamline group creation a little bit: __init__() only
    sets up the shell of a Group, and nothing happens with the obj until it is assigned as
    an attribute of a ParserBase object (which will then assign the group's name and set it
    up to be an actual SubHandler and everything).
    """
    def __init__(self, *, required=False, AND=_Null, OR=_Null, XOR=_Null):
        """
        CLI later calls `SubHandler.__init__(Group(), self, name)` in its __setattr__()
        (All instance attributes are overridden by this)
        """
        self._required = required
        self._and = AND
        self._or = OR
        self._xor = XOR
    
    def arg(self, n=1, **kwargs):
        """
        See _Handler.arg()
        
        n: number of times this arg should be received consecutively; pass ... for infinite
        Expected kwargs: _ (str), help (str)
        """
        def inner(cb):
            entity = Arg(cb, n, namespace=kwargs.get('namespace'), help=kwargs.get('help'))
            self.arg_map[entity.name] = entity
            if kwargs.get('required'):
                self._required.add(entity.name)
            return self.parent.arg(n, **kwargs)(entity.func)
        return inner
    
    def flag(self, dest=None, short=_Null, **kwargs):
        """
        See _Handler.flag().
        """
        def inner(cb):
            entity = Flag(cb, namespace=kwargs.get('namespace'), name=dest, help=kwargs.get('help'), _=kwargs.get('_', '-'))
            if cb.__name__.isidentifier():
                self._aliases[cb.__name__] = entity.name
            if short is not None:  # _Null == default; None == none
                try:
                    entity.short = short or next(s for s in entity.name if s.isalnum() and s not in self._aliases)
                except StopIteration:
                    pass
                else:
                    self._aliases[entity.short] = entity.name
            if kwargs.get('default', _Null) is not _Null:  # could be `'default' not in kwargs` but we don't want them using _Null either
                self._defaults[entity.identifier] = kwargs['default']
            if kwargs.get('required'):
                self._required.add(entity.name)
            self.flags[entity.name] = entity
            return self.parent.flag(dest, short, **kwargs)(entity.func)
        return inner


class Command(SubHandler, ParserBase):
    """
    Subcommand: attached to a parser but is also one in its own right.
    """
    def __init__(self, flag_prefix='-', *, parent, name, desc):
        SubHandler.__init__(self, parent, name)
        ParserBase.__init__(self, desc, flag_prefix, systemexit=getattr(parent, 'systemexit', True))
    
    def __str__(self):
        return ' {}'.format(self.name)
    
    @classmethod
    def from_cli(cls, cli, parent, name):
        """
        cli: joffrey.core.CLI object to create command from
        parent: ParserBase object to attach new command to
        name: name of new command

        Creates a command from a preexisting ParserBase object. Helps with extensibility.
        """
        obj = cls(cli.flag_prefix, parent=parent, name=name, desc=cli.desc)
        obj.__dict__.update(vars(cli))
        return obj
    
    @property
    def help(self):
        return self.format_help()


class CLI(ParserBase):
    """
    The 'main dish', as phrased in the README.
    This is what users import and base their joffrey applications off of.
    """
    def __str__(self):  # for help screen (because main CLI shouldn't show its own name)
        return ''
    
    def _extract_flargs(self, *args, **kwargs):
        # Top-level CLI has nothing to bubble its unknowns up to
        kwargs['propagate_unknowns'] = False
        # Propagate_unknowns is the last pos arg for now, so we can do [:-1] to get rid of it
        return super()._extract_flargs(*args[:-1], **kwargs)



import inspect
from ast import literal_eval
from functools import partial, wraps
from itertools import islice, starmap
from types import SimpleNamespace


_Null = type(
  '_NullType', (),
  {'__bool__': lambda self: False,
   '__repr__': lambda self: '<_Null>'}
  )()
VAR_POSITIONAL, KEYWORD_ONLY = inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.KEYWORD_ONLY


def _callable(obj):
    return callable(obj) and obj is not inspect._empty


def convert(hint, val):
    return hint(val) if _callable(hint) else val


def typecast(func):
    """
    Wraps func such that arguments passed to it will be converted
    according to its typehints.
    
    More specifically, calls func's annotations on arguments
    passed to it; non-callable annotations are not touched.
    If a callable annotates a variadic argument (*, **),
    the annotation will be called on each value therein.
    """
    def _hint_for(param):
        return func.__annotations__.get(param.name)

    params = inspect.signature(func).parameters.values()
    # Gather annotations
    # ...of positional parameters
    pos = [_hint_for(p) for p in params if p.kind < VAR_POSITIONAL]
    var_pos = next((_hint_for(p) for p in params if p.kind == VAR_POSITIONAL), None)
    pos_defaults = [p.default for p in params if p.kind < VAR_POSITIONAL]
    
    # ...of keyword parameters
    kw = {p.name: _hint_for(p) for p in params if p.kind == KEYWORD_ONLY}
    var_kw = next((_hint_for(p) for p in params if p.kind > KEYWORD_ONLY), None)
    kw_defaults = {p.name: p.default for p in params if p.kind == KEYWORD_ONLY}
    
    @wraps(func)
    def wrapper(*args, **kwargs):
        args_, kwargs_ = [], {}
        # Can use a consumable generator to keep track of what
        # positionals are left to convert
        arg_iter = iter(args)
        if len(args) > len(pos) and not var_pos:
            # More positional arguments were passed than func accepts
            func(*args, **kwargs)  # raise TypeError
        # Type-convert the positional arguments that were passed as such
        args_.extend(starmap(convert, zip(pos, arg_iter)))
        # Fill in the rest with either positional parameters passed as kwargs
        # or, failing that, each parameter's default value
        for param, hint, default in islice(zip(params, pos, pos_defaults), len(args_), None):
            if param.name in kwargs:
                args_.append(convert(hint, kwargs.pop(param.name)))
            else:
                args_.append(default)
        # If some positionals aren't present and also don't have defaults,
        if inspect._empty in args_:
            # Then they were simply not passed as positionals,
            # but they may have been passed via keyword:
            for idx, (param, hint, passed) in enumerate(zip(params, pos, args_)):
                if passed is not inspect._empty:
                    # Only look at those for which nothing was passed
                    continue
                try:
                    args_[idx] = convert(hint, kwargs.pop(param.name))
                except KeyError:
                    # Then this parameter wasn't given, period
                    func(*args, **kwargs)  # raise TypeError
        # If func accepts *args and arg_iter has any values left in it, they
        # should be passed to *args
        if var_pos is not None:
            args_.extend(map(var_pos, arg_iter) if _callable(var_pos) else arg_iter)
        # Keyword-parameter typehints:
        for name, hint in kw.items():
            try:
                kwargs_[name] = convert(hint, kwargs[name])
            except KeyError:
                default = kw_defaults[name]
                if default is inspect._empty:
                    # Keyword argument was not passed and has no default
                    func(*args, **kwargs)  # raise TypeError
                kwargs_[name] = default
        # **kwargs: just convert every value while keeping the dict otherwise intact
        if var_kw is not None:
            kwargs_.update({name: convert(var_kw, val) for name, val in kwargs.items() if name not in kwargs_})
        return func(*args_, **kwargs_)
    return wrapper


def booly(arg):
    """
    arg: str representing something boolean-like
    return: boolean representation of `arg`
    """
    comp = arg.lower()
    if comp in ('yes', 'y', 'true', 't', '1'):
        return True
    elif comp in ('no', 'n', 'false', 'f', '0'):
        return False
    else:
        raise ValueError('Could not convert {!r} to boolean'.format(arg))


class auto:
    """
    Performs literal_eval for whatever it's called on,
    optionally checking types
    """
    def __new__(cls, obj, *rest):
        if isinstance(obj, str) and not rest:
            return cls._leval(obj)
        return super().__new__(cls)
    
    def __init__(self, *types):
        self.types = types
        self.negated = False
        for type_ in self.types:
            if not isinstance(type_, type):
                raise TypeError("auto() argument '{}' is not a type".format(type_))
    
    def __invert__(self):
        """
        Toggle whether to check
          isinstance(..., self.types)
        vs
          not isinstance(...)
        """
        self.negated ^= True
        return self
    
    def __call__(self, obj):
        ret = self._leval(obj)
        if self.negated:
            if isinstance(ret, self.types):
                raise TypeError('Did not expect {}-type {!r}'.format(
                  type(ret).__name__,
                  ret
                ))
        elif not isinstance(ret, self.types):
            raise TypeError('Expected {}, got {} {!r}'.format(
              '/'.join(i.__name__ for i in self.types),
              type(ret).__name__,
              ret
            ))
        return ret
    
    @staticmethod
    def _leval(s):
        """
        s: str to evaluate as literal
        return: literal represented by s (or s itself if none)
        """
        try:
            return literal_eval(s)
        except (SyntaxError, ValueError):
            return s


class multiton:
    """
    Decorator that turns a class into a multiton;
    sameness is determined by class.__init__() arguments.
    """
    classes = {}
    
    def __init__(self, pos=None, *, kw=False, cls=None, hash_by=None):
        """
        pos: where to stop reading positional arguments (i.e. read args[:pos])
        kw: whether to take into account keyword arguments
        cls: class to consider this multiton a part of (i.e. class to instantiate
          when user tries to instantiate the decorated class)
        hash_by: a hash function for when args/kwargs are mutable
        """
        self.class_ = cls
        self.kw = kw
        self.pos = pos
        self.hash_func = hash_by
    
    def __call__(self, deco_cls):
        cls = self.class_ or deco_cls
        instances = self.classes.setdefault(cls, {})
        
        def get_instance(*args, **kwargs):
            key = (args[:self.pos], kwargs) if self.kw else args[:self.pos]
            if self.hash_func is not None:
                key = tuple(map(self.hash_func, key))
            if key not in instances:
                instances[key] = deco_cls(*args, **kwargs)
            return instances[key]
        
        get_instance.cls = deco_cls
        return get_instance


class JoffreyNamespace(SimpleNamespace):
    """
    dict+SimpleNamespace with a `_` attribute
    for additional dict-like stuff
    (needed because, say, JoffreyNamespace().values()
    may conflict with an item named 'values'; a single
    underscore is less likely to conflict with anything
    chosen by a user)
    """
    def __bool__(self):
        return bool(vars(self))
    
    def __eq__(self, other):
        return vars(self) == other
    
    def __contains__(self, name):
        return self._._contains_(name)
    
    def __getitem__(self, name):
        return self._._getitem_(name)
    
    def __getattr__(self, name):
        return self._._getitem_(name)
    
    def __iter__(self):
        yield from vars(self)
    
    @property
    def _(self):
        return _SubNamespace(self)


@multiton(hash_by=id)
class _SubNamespace:
    def __init__(self, parent):
        self._parent = parent
        parent_dict = vars(parent)
        self.items = parent_dict.items
        self.keys = parent_dict.keys
        self.values = parent_dict.values
        self.get = parent_dict.get
        
        # These used to be overridable due to the whole 'default key' thing
        self._getitem_ = parent.__getattribute__
        self._contains_ = parent_dict.__contains__
    
    def pretty(self, delim='\n', sep=': '):
        """
        delim: string to join items with
        sep: what to separate key from value with
        """
        return delim.join(
          '{}{}{}'.format(k, sep, v)
          for k, v in self.items()
          )



import setuptools

with open('README.md', encoding='utf-8') as f:
    long_description = ''.join(f.read().split('[](#separator-for-pypi)')[::2])

setuptools.setup(
  name='joffrey',
  license='MIT',
  version='0.5.2',
  author='Hadi Tarhini',
  author_email='eltrhn@gmail.com',
  url='https://github.com/supposedly/joffrey',
  description='A powerful Python 3 command-line parser',
  long_description=long_description,
  long_description_content_type='text/markdown',
  keywords='argparse argument parsing',
  python_requires='>=3.5',
  packages=setuptools.find_packages(),
  classifiers=(
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
  ),
)



import inspect
import sys
from copy import deepcopy

from .misc import multiton, typecast


VAR_POS = inspect.Parameter.VAR_POSITIONAL


@multiton()
class Entity:
    """
    Base class for flags/positional arguments.

    ###  instance attrs  ###
    _namespace: dict containing namespace-initialization values
    _normalized_params: names of arguments this entity takes, formatted prettily
    params: arguments this entity takes
    argcount: how many args this entity can take
    func: original function passed to __init__()
    callback: func decorated with joffrey.misc.typecast()
    help: entity's helptext
    identifier: custom name if __init__() was given one else func.__name__
    name: identical to identifier but subclasses can override
    """
    def __init__(self, func, *, name=None, namespace=None, help=None):
        """
        func: function this entity is to call
        name: entity's name or func.__name__
        namespace: dict with which to initialize a namespace if applicable
        help: entity's help text or func.__doc__
        """
        params = inspect.signature(func).parameters
        has_nsp = bool(namespace)
        first_optional = next((i for i, v in enumerate(params.values()) if v.default is not inspect._empty), sys.maxsize)
        self._namespace = namespace
        self.params = list(params)[has_nsp:]
        self.argcount = sys.maxsize if any(i.kind == VAR_POS for i in params.values()) else len(params) - has_nsp
        self._normalized_params = [
          ('({})'.format(v) if i >= first_optional else v).upper()
          for i, v in enumerate(self.params[:-1])
          ]
        if self.params:
            last = self.params[-1]
            if self.argcount == sys.maxsize:
                self._normalized_params.append('{}...'.format(last).upper())
            else:
                self._normalized_params.append(('({})'.format(last) if len(params) >= first_optional else last).upper())
        self.func, self.callback = func, typecast(func)
        self.help = inspect.cleandoc(func.__doc__ or '' if help is None else help)
        self.brief = next(iter(self.help.split('\n')), '')
        self.identifier = name or func.__name__
        self.name = self.identifier
    
    @property
    def namespace(self):
        # must deepcopy because users can modify the returned dict
        return deepcopy(self._namespace)
    
    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)


@multiton(cls=Entity.cls)
class Flag(Entity.cls):
    """
    A flag AKA option (the kind passed in unix with - and --)
    name: identifier but with its underscores replaced by whatever given to __init__()
    short: short alias
    """
    def __init__(self, *args, _='-', **kwargs):
        """
        _: What to replace underscores with in self.name
        """
        super().__init__(*args, **kwargs)
        self.name = self.identifier.replace('_', _)
        self.short = None  # can be reassigned
    
    @property
    def args(self):
        return ' ' + ' '.join(self._normalized_params) if self.params else ''
    
    def __str__(self):
        if self.short is None:
            return '[--{}{}]'.format(self.name, self.args)
        return '[-{} | --{}{}]'.format(self.short, self.name, self.args)


@multiton(cls=Entity.cls)
class Arg(Entity.cls):
    """
    A positional argument
    name: self.identifier
    repcount: how many times this argument is to be consecutively invoked
    """
    def __init__(self, cb, repeat_count, **kwargs):
        super().__init__(cb, **kwargs)
        self.name = self.identifier
        self.repcount = repeat_count
    
    def __str__(self):
        return '{}({})'.format(self.identifier, '...' if self.repcount is Ellipsis else self.repcount)



import inspect
import shlex
import sys

from .core import CLI
from .misc import _Null, JoffreyNamespace, convert


class Simpleton:
    _ = '-'
    flag_prefix = '-'
    short_flags = True
    no_help = False
    
    def __init__(self, func, cli=None):
        self.cli = cli or CLI(no_help=self.no_help, flag_prefix=self.flag_prefix)
        # Fossilize these so if they're changed on the class it doesn't modify this instance
        self._ = self._
        self.flag_prefix = self.flag_prefix
        self.short_flags = self.short_flags
        self.no_help = self.no_help
        
        self._consuming = False
        self.commands = {}
        
        self._callback = func
        self._params = inspect.signature(func).parameters.values()
        self.cli.desc = func.__doc__ or ''
        
        pos = [i for i in self._params if i.kind <= inspect.Parameter.VAR_POSITIONAL]
        flags = [i for i in self._params if i.kind == inspect.Parameter.KEYWORD_ONLY]
        
        self._consuming = pos and pos[-1].kind == inspect.Parameter.VAR_POSITIONAL
        self._add_flargs(self.cli, pos, flags)
    
    def __call__(self, *args, **kwargs):
        return self._callback(*args, **kwargs)
    
    @staticmethod
    def _null_check(val):
        return val is _Null or val is inspect._empty
    
    @classmethod
    def no_top_level(cls, help=''):
        new = cls(lambda: None)
        new.cli.desc = help
        return new
    
    def _add_flargs(self, cli, pos, flags):
        for arg in pos:
            if arg.kind == inspect.Parameter.VAR_POSITIONAL:
                def __hidden(arg):
                    def __inner(nsp, value):
                        nsp.accum.append(convert(arg.annotation, value))
                        return nsp.accum
                    __inner.__name__ = arg.name
                    return __inner
                cli.arg(
                  Ellipsis,
                  namespace={'accum': []},
                  required=False,
                  default=_Null if self._null_check(arg.default) else arg.default,
                )(__hidden(arg))
            else:
                def __hidden(arg):
                    def __inner(value):
                        return convert(arg.annotation, value)
                    __inner.__name__ = arg.name
                    return __inner
                cli.arg(
                  required=self._null_check(arg.default),
                  default=_Null if self._null_check(arg.default) else arg.default,
                )(__hidden(arg))
        
        for flag in flags:
            def __hidden(flag):
                def __inner(value):
                    return convert(flag.annotation, value)
                __inner.__name__ = flag.name
                return __inner
            cli.flag(
              short=_Null if self.short_flags else None,
              default=_Null if self._null_check(flag.default) else flag.default,
              _=self._
            )(__hidden(flag))
    
    def command(self, func):
        self.commands[func.__name__] = cmd = Simpleton(func, self.cli.command(func.__name__, _=self._))
        cmd.no_help = self.no_help
        cmd.short_flags = self.short_flags
        cmd._ = self._
        return cmd
    
    def call(self, **flags):
        args = []
        commands = []
        for name, val in flags.copy().items():
            if isinstance(val, JoffreyNamespace):
                # This used to be:
                # self.commands[name].call(**dict(flags.pop(name)._.items()))
                # But commands being run before the main callback was bad
                # So now it's deferred until the `for cmd, flargs in commands:` loop
                commands.append((self.commands[name], dict(flags.pop(name)._.items())))
            elif name in self._params:
                args.append(flags.pop(name))
        
        args = [
          flags.pop(p.name, ())
            if p.kind == inspect.Parameter.VAR_POSITIONAL
          else flags.pop(p.name)
            if p.name in flags or p.default is inspect._empty
          else p.default
            for p in self._params
            if p.kind <= inspect.Parameter.VAR_POSITIONAL
          ]
        
        if self._consuming:
            args.extend(args.pop(-1))
        
        ret = self._callback(*args, **flags)
        
        for cmd, flargs in commands:
            cmd.call(**flargs)
        return ret
    
    def run(self, inp=None):
        return self.call(**dict(self.cli.parse(inp)._.items()))
    
    def search(self, inp=None):
        if inp is None:
            inp = sys.argv[1:]
        if isinstance(inp, str):
            inp = shlex.split(inp)
        if not isinstance(self.cli, CLI):
            try:
                idx = next(
                  i for i, v in enumerate(inp)
                  if v == self._callback.__name__
                  and not
                  (inp[i-1].startswith(self.cli.flag_prefix) or '=' in inp[i-1].lstrip(self.cli.flag_prefix))
                  )
            except StopIteration:
                raise IndexError('command invocation {!r} not found in input'.format(self._callback.__name__))
            inp = inp[idx:]
        return self.run(inp)


sys.modules[__name__] = Simpleton
